
//#define DEBUG

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <limits.h>
#include <map>
#include <list>
#include <vector>
#include <stdlib.h>
#include <assert.h>

#include <unordered_map>
#include <unordered_set>
#include <string>



static const uint64_t NUM_CHILDREN = 16;

                
class Sigma_Tree
{
public:
    Sigma_Tree(uint64_t _num_timestamps);

    void Increment(uint64_t from, uint64_t to);
    void Decrement(uint64_t from, uint64_t to);
    unsigned Get_Max(uint64_t from, uint64_t to);
    unsigned Vertical_Sum(uint64_t from, unsigned level);

private:
    std::vector<std::vector<unsigned> > value_tree_data;
    std::vector<std::vector<unsigned> > max_tree_data;
    uint64_t                num_timestamps;

    unsigned Utility(uint64_t from, uint64_t to, 
                    int delta, unsigned level);
    void Vertical_Max(uint64_t from, unsigned level, int delta);
};

Sigma_Tree::Sigma_Tree(uint64_t _num_timestamps)
{
    uint64_t addition = num_timestamps = _num_timestamps;
    uint64_t last = ~0;

    while (addition != last) {
        assert(addition > 0);

        std::vector<unsigned> v1;
        std::vector<unsigned> v2;
        v1.resize(addition);
        value_tree_data.push_back(v1);

        if (!max_tree_data.empty()) {
            v2.resize(addition);
        }
        max_tree_data.push_back(v2);

        last = addition;
        addition = (addition + NUM_CHILDREN - 1) / NUM_CHILDREN;
    }
}

unsigned Sigma_Tree::Get_Max(uint64_t from, uint64_t to)
{
    if (from <= to) {
        return Utility(from, to, 0, 0);
    }
    return 0;
}

void Sigma_Tree::Increment(uint64_t from, uint64_t to)
{
    if (from <= to) {
        Utility(from, to, 1, 0);
    }
}

void Sigma_Tree::Decrement(uint64_t from, uint64_t to)
{
    if (from <= to) {
        Utility(from, to, -1, 0);
    }
}

unsigned Sigma_Tree::Vertical_Sum(uint64_t from, unsigned level)
{
    unsigned total = 0;
    
    if (level > 0) {
        total = max_tree_data[ level ][ from ];
    }

    while (level < value_tree_data.size()) {
        std::vector<unsigned> & value_level_data = value_tree_data[ level ];
        total += value_level_data[ from ];
        level ++;

        from /= NUM_CHILDREN;
        if (level < value_tree_data.size()) {
            assert (total <= max_tree_data[ level ][ from ]);
        }
    }
    return total;
}

void Sigma_Tree::Vertical_Max(uint64_t from, unsigned level, int delta)
{
    // New value inserted
    unsigned value = value_tree_data[ level ][ from ] + delta;
    value_tree_data[ level ][ from ] = value;

    // printf("inc %u %u to %u\n", level, (unsigned) from, value);

    if (level > 0) {
        value += max_tree_data[ level ][ from ];
    }

    // Propagate new maximum if necessary
    while (true) {
        level ++;
        from /= NUM_CHILDREN;

        if (level >= value_tree_data.size()) {
            break;  // reached root
        }
        unsigned max = max_tree_data[ level ][ from ];
        
        if (((delta > 0) && (value <= max))
        || ((delta < 0) && (value >= max))) {
            break; // no new maximum
        } else {
            // printf(" max %u %u is %u\n", level, (unsigned) from, value);
            max_tree_data[ level ][ from ] = value;
        }
        value += value_tree_data[ level ][ from ];
    }
}

unsigned Sigma_Tree::Utility(uint64_t from, uint64_t to, 
                    int delta, unsigned level)
{
    unsigned max_so_far = 0;
    uint64_t target, i;
    std::vector<unsigned> & value_level_data = value_tree_data[ level ];

    assert (from <= to);
    assert (to < value_level_data.size());

    // From the left
    target = ((from + NUM_CHILDREN - 1) / NUM_CHILDREN) * NUM_CHILDREN;
    while ((from < to) && (from < target)) {
        if (delta != 0) {
            Vertical_Max(from, level, delta);
        } else {
            unsigned value = Vertical_Sum(from, level);
            if (value > max_so_far) {
                max_so_far = value;
            }
        }
        from ++;
    }
    // From the right
    target = (to / NUM_CHILDREN) * NUM_CHILDREN;
    if (target < from) {
        target = from;
    }
    for (i = target; i <= to; i++) {
        if (delta != 0) {
            Vertical_Max(i, level, delta);
        } else {
            unsigned value = Vertical_Sum(i, level);
            if (value > max_so_far) {
                max_so_far = value;
            }
        }
    }
    to = target;

    // Next level
    level ++;
    uint64_t next_from = from / NUM_CHILDREN;
    uint64_t next_to = to / NUM_CHILDREN;
   
    if (next_from < next_to) {
        assert (level < value_tree_data.size());
        unsigned value = Utility(next_from, next_to - 1, delta, level);
        if (value > max_so_far) {
            max_so_far = value;
        }
    }
    return max_so_far;
}


int main(int argc, char ** argv)
{
    if (argc != 4) {
        printf("usage: belady "
            "<trace_in file> <cache size> <line size>\n"
            "<cache size>           in bytes, e.g. 8192.\n"
            "<line size>            in bytes, e.g. 4.\n");
        return 1;
    }
    FILE * trace_in = fopen(argv[ 1 ], "rb");
    if (trace_in == NULL) {
        printf("unable to open trace_in file\n");
        return 1;
    }

    unsigned memory_size = atoi(argv[ 2 ]);
    unsigned line_size = atoi(argv[ 3 ]);
    if ((memory_size == 0)
    || (line_size == 0)
    || ((memory_size / line_size) > (1 << 24))
    || ((memory_size / line_size) == 0)) {
        printf("sizes are invalid\n");
        return 1;
    }

    memory_size /= line_size;

    // how many addresses?

    fseek(trace_in, 0, SEEK_END);
    long end_of_file = ftell(trace_in);
    assert(end_of_file > 0);
    assert((end_of_file & 3) == 0);
    
    uint64_t num_addresses = end_of_file / 4;
    const uint64_t OFFSET = 10;
    uint64_t c_miss_count = 0;
    uint64_t miss_count = 0;
    uint64_t hit_count = 0;
    uint64_t transfer_count = 0;
    uint64_t i;
    std::unordered_map<unsigned, uint64_t> address_to_last_usage;
    uint8_t * store_in_cache = new uint8_t[(num_addresses >> 3) + 1];

    Sigma_Tree sigma_tree(num_addresses + OFFSET);

    // Run OPT algorithm on address sequence
    fseek(trace_in, 0, SEEK_SET);
    for (i = 0; i < num_addresses; i++) {
        uint64_t sT = 0;
        uint64_t current_time = i + OFFSET;
        uint32_t address;

        fread(&address, 4, 1, trace_in);
        address /= line_size;
        
        if (address_to_last_usage.count(address)) {
            sT = address_to_last_usage[ address ];

            unsigned max = sigma_tree.Get_Max(sT, current_time - 1);
            if (max >= memory_size) {
                sT = 0; // Can't be allocated
            }
        } else {
            c_miss_count ++;
        }

        address_to_last_usage[ address ] = current_time;
        if (sT) {
            sigma_tree.Increment(sT, current_time - 1);

            uint64_t bit = sT - OFFSET;
            store_in_cache[bit >> 3] |= 1 << (bit & 0x7);
        } 
    }
    unsigned max = sigma_tree.Get_Max(0, num_addresses - 1);
    assert (max <= memory_size);

    // Actually simulate the cache based on the replacement
    // decisions made in the earlier stage
    std::unordered_map<unsigned, uint64_t> cache_contents;
    std::unordered_map<unsigned, uint64_t>::iterator ci;

    fseek(trace_in, 0, SEEK_SET);
    for (i = 0; i < num_addresses; i++) {
        uint32_t address;

        fread(&address, 4, 1, trace_in);
        address /= line_size;

        if (cache_contents.count(address)) {
            // A hit
            hit_count ++;

            if (!(store_in_cache[i >> 3] & (1 << (i & 0x7)))) {
                // Data evicted from cache now
                cache_contents.erase(address);
            }
        } else {
            // A miss
            miss_count ++;

            if ((store_in_cache[i >> 3] & (1 << (i & 0x7)))) {
                // Data moved into cache now
                cache_contents[ address ] = i;
                transfer_count ++;
            }
        }
        assert(cache_contents.size() <= memory_size);
    }
    fclose(trace_in);

    cache_contents.clear();

    printf("%-8u %-8u %-8.0f %-8.0f %-8.0f %s\n", 
        memory_size * line_size,
        line_size,
        (double) hit_count,
        (double) miss_count,
        (double) transfer_count,
        argv[1]);
    fflush(stdout);
}


