#!/usr/bin/python

import os, glob, sys, pickle, struct, collections
import treeise

BUILD_ONLY = False
CLEAN = True
ARM_ELF = "out-arm.elf"
PROGRAMS_DIR = "programs"
LDFLAGS = '-lm -lstdc++ -lgcc'
PREFIX = "arm-none-linux-gnueabi-"
M5DIR = "../../gem5/gem5-stable"
M5BIN = "build/ARM_SE/m5.opt"
SEBIN = "configs/example/se.py"
TPI = 500
MICRO_OPS = 50e7
RUN_MASK = 0xffffffff

def csystem(name):
    print name
    rc = os.system(name)
    if rc != 0:
        print 'system call failed: %s' % name
        sys.exit(1)

def copy(src, dest):
    assert os.path.isfile(src)
    assert os.path.isfile(dest) or not os.path.exists(dest)
    file(dest, "wb").write(file(src, "rb").read())

def build(program, args, build_for_call_graph):
    args = [program] + args

    m5dir = os.path.abspath(M5DIR)
    original = os.getcwd()
    build_dir = os.path.join(original, PROGRAMS_DIR, program)

    if not os.path.isdir(PROGRAMS_DIR):
        os.mkdir(PROGRAMS_DIR)
    if not os.path.isdir(build_dir):
        os.mkdir(build_dir)
    if os.path.isfile(os.path.join(build_dir, ARM_ELF)):
        print build_dir, 'already done'
        return

    os.environ[ "R" ] = root = build_dir
    os.chdir(build_dir)

    symlinks = [program, "m5trace"]
    for p in symlinks:
        os.symlink(os.path.join(original, p), p)

    if build_for_call_graph:
        bitcodes = []
        for c in args:
            assert c.endswith(".c")
            c = c[:-2]
            csystem("llvm-gcc -emit-llvm -Os -o %s.bc -c %s.c" % (c, c))
            bitcodes.append("%s.bc" % c)

        csystem("llvm-ld -b out1.bc -disable-opt -disable-inlining %s" % ' '.join(bitcodes))

        csystem("opt -f -loopsimplify -o out0.bc out1.bc")
        csystem("valgrind opt -f -loop-extract -o out1.bc out0.bc")

        csystem("llc -f -o out.S out1.bc --disable-fp-elim -O2 "
                "-march=arm -mcpu=cortex-a8")
        csystem("%sas -o out.o out.S -mcpu=cortex-a8 -mfpu=neon-vfpv4" % (PREFIX))
        csystem("%sgcc -static -o %s out.o %s" % (PREFIX, ARM_ELF, LDFLAGS))

    else:
        csystem("%sgcc -static -o %s -O3 %s %s" % (PREFIX, ARM_ELF,
                    LDFLAGS, ' '.join(args)))

    csystem("%snm -aS %s >out.nm" % (PREFIX, ARM_ELF))
    csystem("%sobjdump -d %s >out.od" % (PREFIX, ARM_ELF))

    main_at = None
    for line in file("out.nm"):
        line = line.rstrip()
        if line.endswith(" T main"):
            main_at = int(line.split()[0], 16)

    if main_at == None:
        raise Exception('could not find the main function')

    fifo_name = os.path.abspath("f1")
    file("./runme.sh", "wt").write(file(os.path.join(original, "groove.dat")).read() % {
        "trace1" : fifo_name,
        "trace2" : os.path.abspath("m5trace."),
        "m5dir" : m5dir,
        "m5bin" : "%s/%s" % (m5dir, M5BIN),
        "sebin" : "%s/%s" % (m5dir, SEBIN),
        "startticks" : TPI * 0,
        "endticks" : TPI * MICRO_OPS,
        "tracebin" : "%s/m5trace" % root,
        "main_at" : main_at,
        "arm" : os.path.abspath(ARM_ELF),
    })
    os.chmod("./runme.sh", 0755)

    csystem("./runme.sh")

    for p in symlinks + [fifo_name]:
        os.unlink(p)

def pinch(root, out):
    try:
        arrangement = out["arrangement"]
        program = out["program"]
        spm_size = out["spm_size"]
        fixed_cost = out["fixed_cost"]
    except KeyError:
        return
        
    target = os.path.join(root, "tables", 
            arrangement, program, str(spm_size), str(fixed_cost))
    csystem("mkdir -p %s" % target)

    for interesting in INTERESTING:
        if os.path.exists(interesting):
            copy(interesting, os.path.join(target, interesting))

    #copy(ARM_ELF, os.path.join(target, ARM_ELF))
    pickle.dump(out, file("%s/out.dat" % target, "wb"))
   
def wipe():
    for interesting in INTERESTING:
        if os.path.exists(interesting):
            os.unlink(interesting)

INTERESTING = [ "tr.2", "gr.2", "gr.3",
        "out.cs", "g.gv", "g2.tr", "out.dm", "g2.gv",
        ]

def collect(spm_size = 256, fixed_cost = 20, mode = "W"):
    trace_in = "m5trace.1"
    graph_out = "tr.2"
    csystem("$R1/tree_get %s %s %s %u" % (trace_in, 
                "out.nm", graph_out, spm_size))

    lookup = get_lookup()
    g = treeise.Graph()
    g.read_python(graph_out)
    g.finalise(False)
    assert g.is_tree()

    g.write_graphviz("g.gv")
    csystem("dot -Tpdf -og.pdf g.gv")

    g.remove_leaf()

    g.write_graphviz("g.gv")
    csystem("dot -Tpdf -og.pdf g.gv")

    g.write_csearch("out.cs", spm_size = spm_size, 
            mode = mode, quantum = 4, fixed_cost = fixed_cost)

    import csp
    check_value = csp.do_csearch("out.cs", "L")
    assert check_value != 0
    assert check_value < 1e9

    g.read_csearch("out.cs")
    g.write_graphviz("g2.gv")
    g.write_python("g2.tr")
    csystem("dot -Tpdf -og2.pdf g2.gv")

    #for m in g.methods:
    #    print '%5u %s' % (m.size, m.name)
    #walk(g.root, "", lookup, 1, file("out.soln", "wt"))

    region_size = g.get_region_sizes()

    child = g.root
    total = 0
    cost = 1
    if False:
        print 'entry cost %d * (%d + fc) = %d' % (
                        cost, region_size[child.region],
                        cost * (region_size[child.region] + fixed_cost))
        total += cost * region_size[child.region] + fixed_cost
        for parent in g.methods:
            for (child, cost) in parent.tree_children:
                if parent.region != child.region:
                    print 'region transition edge R%d -> R%d at m%d:' % (
                                    parent.region, child.region,
                                    child.number),
                    print 'cost %d * (%d + fc) = %d' % (
                                    cost, region_size[child.region],
                                    cost * (region_size[child.region] + fixed_cost))
                    total += cost * (region_size[child.region] + fixed_cost)

            for (footer, cost) in parent.loop_from:
                if footer.region != parent.region:
                    child = footer
                    total += cost * (region_size[parent.region] + fixed_cost)

        print total
    #output_program(g, file("g2.prog", "wt"), lookup)
    tmp_trace = file("/tmp/a.log", "wt")
    spm_cost = follow_program(g, tmp_trace, lookup, fixed_cost)
    tmp_trace = os.tmpfile()
    spm_bytes = follow_program(g, tmp_trace, lookup, 0)

    cache_cost = []
    cache_bytes = []
    for line_size in [4, 8, 16, 32]:
        if line_size <= spm_size:
            csystem("$R1/directmap m5trace.1 %u %u >out.dm" % (
                        spm_size, line_size))
            cache_misses = (int(file("out.dm").read()))
            cache_bytes.append((cache_misses * line_size, line_size))
            cache_cost.append((cache_misses * (line_size + fixed_cost), 
                            line_size))

    cache_cost.sort()
    cache_bytes.sort()

    value = g.get_value(fixed_cost)
    assert value <= check_value

    return {
        "spm_size" : spm_size,
        "fixed_cost" : fixed_cost,
        "spm_bytes" : spm_bytes,
        "spm_cost" : value,
        "num_regions" : len(region_size),
        "arrangement" : "ELA2",
        "cache_cost" : cache_cost[0][0],
        "cache_bytes" : cache_bytes[0][0],
        "cache_cost_ls" : cache_cost[0][1],
        "cache_bytes_ls" : cache_bytes[0][1],
    }

def follow_program(g, out, lookup, fixed_cost):
    
    trace = file("m5trace.1", "rb")
    replay = collections.deque()

    def read_next():
        if len(replay) != 0:
            return replay.popleft()

        word = trace.read(4)
        (address, ) = struct.unpack("<L", word)
        address &= ~3
        return address
    
    def unread(address):
        replay.append(address)

    labels = dict()
    region_size = g.get_region_sizes()
    current_region = [99999]

    def dump_method(m):
        total = None
        try:
            total = _dump_method(m)
        finally:
            if total == None:
                out.write("unstack number %d\n" % m.number)
                total = 0

        return total

    def _dump_method(m):
        total = 0
        out.write("\n")
        out.write("following @ node (%u) %u\n" % (m.number, trace.tell()))
        if current_region[0] != m.region:
            total += region_size[m.region] + fixed_cost
            out.write("transition R%d -> R%d at m%d incurs cost %d\n" % (
                    current_region[0], m.region, m.number, total))
            current_region[0] = m.region

        is_if = (len(m.tree_children) > 1)
        is_loop = False
        is_call = False

        if len(m.loop_from) != 0:
            for (footer, cost) in m.loop_from:
                out.write("%08x: label (from %08x)\n" % (m.start,
                            footer.start + footer.size))
                labels[footer.start] = m.start

            is_if = False
            if len(m.tree_children) == 1:
                # loop
                is_loop = True
            
            elif len(m.tree_children) == 2:
                is_call = True
                labels[footer.start] += 4

            else:
                assert False, (m.number, len(m.tree_children))

        if is_if:
            out.write("%08s: is if\n" % m.number)
        elif is_loop:
            out.write("%08s: is loop header\n" % m.number)
        elif is_call:
            out.write("%08s: is method call\n" % m.number)
        else:
            out.write("%08s: is ...?\n" % m.number)
                
        for address in xrange(m.start, m.start + m.size, 4):
            assert lookup.has_key(address & RUN_MASK), hex(address)
            out.write("%08x: %s\n" % (address, 
                            lookup[address & RUN_MASK]))
            a0 = read_next()
            if a0 != (address & RUN_MASK):
                out.write("Error: actual address is %08x\n" % a0)
                raise Exception()

        if labels.has_key(m.start):
            out.write("%08s: done? goto %08x\n" % (m.number,
                                    labels[m.start]))

        address = m.start + m.size
        xc = m.tree_children[:]
        xc.sort(key = lambda (child, cost): child.number)

        if is_loop:
            for (child, cost) in xc:
                out.write("%08s: loop %08x\n" % (m.number, child.start))

            looping = True
            while looping:
                # run the first child
                (child, cost) = xc[0]
                total += dump_method(child)

                # where are we now?
                a0 = read_next()
                if a0 == (m.start & RUN_MASK):
                    out.write("%08s: loop continues %08x\n" % (
                                    m.number, m.start))
                    looping = True
                else:
                    out.write("%08s: loop exit %08x (continue is %08x)\n" % (
                                    m.number, a0, m.start))
                    unread(a0)
                    looping = False

        else:
            if labels.has_key(m.start):
                out.write("%08s: loop to %08x\n" % (m.number, labels[m.start]))

            for (child, cost) in xc:
                out.write("%08s: branch %08x\n" % (m.number, child.start))

            a0 = read_next()
            found = 0
            for (child, cost) in xc:
                if (child.start & RUN_MASK) == a0:
                    found += 1

            if labels.has_key(m.start):
                if (labels[m.start] & RUN_MASK) == a0:
                    found += 1

            if found == 0:
                out.write("%08s: no branch candidates "
                        "found, expecting %08x\n" % (m.number, a0))
                raise Exception()

            if found != 1:
                out.write("%08s: ambiguous branch candidates "
                        "found, expecting %08x\n" % (m.number, a0))
                raise Exception()

            for (child, cost) in xc:
                if (child.start & RUN_MASK) == a0:
                    out.write("%08s: taken branch to %08x\n" % (
                                    m.number, child.start))
                    unread(a0)
                    total += dump_method(child)
                    break


        return total

    total = dump_method(g.root)
    out.write("traced to %u; total %u\n" % (trace.tell(), total))
    trace.close()
    return total

def get_lookup():
    lookup = dict()
    for line in file("out.od"):
        fields = line.split()
        if not ((len(fields) >= 3)
        and (fields[0][-1] == ':')
        and (len(fields[1]) == 8)):
            continue

        try:
            address = int(fields[0][:-1], 16)
        except ValueError:
            continue

        lookup[address] = ' '.join(fields[2:])

    return lookup

def objdump_label_factory(lookup):
    def method_label(m):
        out = [ r"(%u) %s R%d" % (m.number, m.name, m.region) ]
        for address in xrange(m.start, m.start + m.size, 4):
            if lookup.has_key(address):
                out.append(r"%x: %s" % (address, lookup[address]))

        return r'\n'.join(out)

    return method_label

def merge(original, out):
    results = os.path.join(original, "results.dat")
    try:
        records = pickle.load(file(results, "rb"))
    except IOError:
        records = []

    records.append(out)
    pickle.dump(records, file(results, "wb"))

    print ''
    print ''
    for (key, data) in sorted(out.items()):
        print '%-20s %s' % (key, data)

    print ''
    print ''

def callgraph(spm_size = 256, fixed_cost = 20):
    trace_in = "m5trace.1"
    graph_out = "gr.2"
    csystem("$R1/graph_get %s %s %s" % (trace_in, "out.nm", graph_out))

    lookup = get_lookup()
    g = treeise.Graph()
    g.read_python(graph_out)
    g.finalise(False)
    g.treeise()
    assert g.is_tree()
    g.write_graphviz("g.gv")
    csystem("dot -Tpdf -og.pdf g.gv")

    for m in g.methods:
        if m.size > spm_size:
            # Impossible
            return {
                "spm_size"    : spm_size,
                "fixed_cost"  : fixed_cost,
                "arrangement" : "ELA1",
                "failure"     : "method of size %u" % m.size,
            }

    g.write_csearch("out.cs", spm_size = spm_size, 
            mode = "C", quantum = 4, fixed_cost = fixed_cost)

    import csp
    check_value = csp.do_csearch("out.cs", "L")
    assert check_value != 0

    g.read_csearch("out.cs")
    g.write_graphviz("g2.gv")
    g.write_python("g2.tr")
    csystem("dot -Tpdf -og2.pdf g2.gv")

    spm_bytes = 0

    region_size = g.get_region_sizes()

    def total_transfer(parent):
        total = 0
        for (child, cost) in parent.tree_children:
            if child.region != parent.region:
                total += cost * (region_size[parent.region] + 
                            region_size[child.region])

            total += total_transfer(child)
        
        return total

    spm_bytes = total_transfer(g.methods[0])
    spm_bytes += region_size[g.methods[0].region]

    return {
        "spm_size"    : spm_size,
        "fixed_cost"  : fixed_cost,
        "spm_bytes"   : spm_bytes,
        "spm_cost"    : check_value,
        "num_regions" : len(region_size),
        "arrangement" : "ELA1",
    }

def twostep(spm_size = 256, fixed_cost = 20):
    trace_in = "m5trace.1"
    graph_out = "gr.3"
    csystem("$R1/graph_get %s %s %s" % (trace_in, "out.nm", graph_out))

    trace = file(trace_in, "rb")
    counter = dict()

    word = trace.read(4)
    while len(word) == 4:
        (address, ) = struct.unpack("<L", word)
        address &= ~3
        counter[address] = counter.get(address, 0) + 1
        word = trace.read(4)

    lookup = get_lookup()
    g = treeise.Graph()
    g.read_python(graph_out)
    g.finalise(False)
    g.treeise()
    assert g.is_tree()
    g.write_graphviz("g.gv")
    csystem("dot -Tpdf -og.pdf g.gv")

    spm_cost = 0
    spm_bytes = 0
    for m in g.methods:
        m.region = m.number

        importance = []
        for address in xrange(m.start, m.start + m.size, 4):
            frequency = counter.get(address, 0)
            importance.append((frequency, address))

        importance.sort()

        m_frequency = counter[m.start]
        alloc_size = 0
        i = 0
        print len(importance)
        while (len(importance) != 0) and (alloc_size <= spm_size):
            (frequency, address) = importance.pop() # in SPM

            old_fetch_time = (fixed_cost + alloc_size) * m_frequency
            fetch_time = (fixed_cost + (alloc_size + 4)) * m_frequency
            fetch_time_delta = fetch_time - old_fetch_time  # from SPM
            run_time_delta = (fixed_cost + 4) * frequency   # from extmem

            if run_time_delta < fetch_time_delta:
                # it's costing me more to fetch this than to use it
                importance.append((frequency, address))
                break

            alloc_size += 4

        while len(importance) != 0:
            (frequency, address) = importance.pop()
            spm_cost += (fixed_cost + 4) * frequency
            spm_bytes += 4 * frequency

        assert counter.get(m.start, 0) > 0
        m.size = alloc_size

    spm_cost += counter[g.methods[0].start] + fixed_cost
    spm_bytes += counter[g.methods[0].start] 

    for parent in g.methods:
        for (child, cost) in parent.tree_children:
            spm_cost += counter[child.start] * (
                    parent.size + child.size + (fixed_cost * 2))
            spm_bytes += counter[child.start] * (
                    parent.size + child.size)

    return {
        "spm_size"    : spm_size,
        "fixed_cost"  : fixed_cost,
        "spm_bytes"   : spm_bytes,
        "spm_cost"    : spm_cost,
        "num_regions" : len(g.methods),
        "arrangement" : "2S",
    }


    
def run(program, args = [], mode = "W"):
    size_list = [64, 128, 256, 512, 1024, 2048, 4096]
    size_list.reverse()

    original = os.getcwd()
    build(program, args, True)
    os.chdir(original)

    if BUILD_ONLY:
        return

    for fixed_cost in [0, 10, 20, 50, 100]:
        os.chdir(os.path.join(original, PROGRAMS_DIR, program))

        for spm_size in size_list:
            wipe()
            out = collect(spm_size = spm_size, 
                    mode = mode, fixed_cost = fixed_cost)
            out["program"] = program
            merge(original, out)
            pinch(original, out)

        for spm_size in size_list:
            wipe()
            out = twostep(spm_size = spm_size, fixed_cost = fixed_cost)
            out["program"] = program
            merge(original, out)
            pinch(original, out)

        for spm_size in size_list:
            wipe()
            out = callgraph(spm_size = spm_size, fixed_cost = fixed_cost)
            out["program"] = program
            merge(original, out)
            pinch(original, out)

        wipe()

    os.chdir(original)



def main():
    # Build and setup
    root = os.getcwd()
    assert os.path.exists(os.path.join(root, "belady.cc"))

    os.environ["R1"] = root
    csystem("make -f binaries.make all")    # C++ programs
    csystem("python csp_setup.py build")    # Python/Pyrex wrapper
    csystem("ln -sf build/*/csp.so")

    import csp                              # Test of C++ code from Python
    test = csp.do_csearch
    assert test != None

    if CLEAN:
        csystem("rm -rf tables results.dat")

    # And at last the actual software
    run("fir.c", mode = "W")  
    run("crc.c", mode="W") 
    run("binarysearch.c") 
    run("matmult.c", mode="w")
    run("edn.c", mode = "w")  
    run("insertsort.c")
    run("jfdctint.c")
    run("bsort100.c") 

if __name__ == "__main__":
    main()

